@@ -60,6 +60,9 @@ ComposerPageState.prototype = {
         }
 
         this.signaturePart = document.getElementById("geary-signature");
+        if (this.signaturePart != null) {
+            ComposerPageState.linkify(this.signaturePart);
+        }
         this.quotePart = document.getElementById("geary-quote");
 
         // Should be using 'e.key' in listeners below instead of
@@ -93,6 +96,12 @@ ComposerPageState.prototype = {
             }
         }, true);
 
+        // Handle file drag & drop
+        document.body.addEventListener("drop", state.handleFileDrop, true);
+        document.body.addEventListener("allowDrop", function(e) {
+            ev.preventDefault();
+        }, true);
+
         // Search for and remove a particular styling when we quote
         // text. If that style exists in the quoted text, we alter it
         // slightly so we don't mess with it later.
@@ -283,38 +292,32 @@ ComposerPageState.prototype = {
     },
     cleanContent: function() {
         // Prevent any modification signals being sent when mutating
-        // the document below.
+        // the document
         this.stopBodyObserver();
-
-        ComposerPageState.cleanPart(this.bodyPart, false);
         ComposerPageState.linkify(this.bodyPart);
-
-        this.signaturePart = ComposerPageState.cleanPart(this.signaturePart, true);
-        this.quotePart = ComposerPageState.cleanPart(this.quotePart, true);
+        this.startBodyObserver();
     },
-    getHtml: function() {
-        // Clone the message parts so we can clean them without
-        // modifiying the DOM, needed when saving drafts. In contrast
-        // with cleanContent above, we don't remove empty elements so
-        // they still exist when restoring from draft
+    getHtml: function(removeEmpty) {
+        if (removeEmpty === undefined) {
+            removeEmpty = true;
+        }
+
         let parent = document.createElement("DIV");
         parent.appendChild(
-            ComposerPageState.cleanPart(this.bodyPart.cloneNode(true), false)
+            ComposerPageState.cleanPart(this.bodyPart.cloneNode(true))
         );
 
-        if (this.signaturePart != null) {
+        if (this.signaturePart != null &&
+            (!removeEmpty || this.signaturePart.innerText.trim() != "")) {
             parent.appendChild(
-                ComposerPageState.cleanPart(
-                    this.signaturePart.cloneNode(true), false
-                )
+                ComposerPageState.cleanPart(this.signaturePart.cloneNode(true))
             );
         }
 
-        if (this.quotePart != null) {
+        if (this.quotePart != null &&
+            (!removeEmpty || this.quotePart.innerText.trim() != "")) {
             parent.appendChild(
-                ComposerPageState.cleanPart(
-                    this.quotePart.cloneNode(true), false
-                )
+                ComposerPageState.cleanPart(this.quotePart.cloneNode(true))
             );
         }
 
@@ -380,6 +383,31 @@ ComposerPageState.prototype = {
             }
         }
         return inPart;
+    },
+    handleFileDrop: function(dropEvent) {
+        dropEvent.preventDefault();
+
+        for (var i = 0; i < dropEvent.dataTransfer.files.length; i++) {
+            const file = dropEvent.dataTransfer.files[i];
+
+            if (!file.type.startsWith('image/'))
+                continue;
+
+            const reader = new FileReader();
+            reader.onload = (function(filename, imageType) { return function(loadEvent) {
+                // Remove prefixed file type and encoding type
+                var parts = loadEvent.target.result.split(",");
+                if (parts.length < 2)
+                    return;
+
+                window.webkit.messageHandlers.dragDropReceived.postMessage({
+                    fileName: encodeURIComponent(filename),
+                    fileType: imageType,
+                    content: parts[1]
+                });
+            }; })(file.name, file.type);
+            reader.readAsDataURL(file);
+        }
     }
 };
 
@@ -387,7 +415,6 @@ ComposerPageState.prototype = {
  * Determines if any keywords are present in a string.
  */
 ComposerPageState.containsKeywords = function(line, wordKeys, suffixKeys) {
-    let urlRegex = ComposerPageState.URL_REGEX;
     let lastToken = -1;
     let lastSpace = -1;
     for (var i = 0; i <= line.length; i++) {
@@ -434,17 +461,12 @@ ComposerPageState.containsKeywords = function(line, wordKeys, suffixKeys) {
 };
 
 /**
- * Removes internal attributes from a composer part..
+ * Removes internal attributes from a composer part.
  */
-ComposerPageState.cleanPart = function(part, removeIfEmpty) {
+ComposerPageState.cleanPart = function(part) {
     if (part != null) {
         part.removeAttribute("class");
         part.removeAttribute("contenteditable");
-
-        if (removeIfEmpty && part.innerText.trim() == "") {
-            part.parentNode.removeChild(part);
-            part = null;
-        }
     }
     return part;
 };
@@ -576,33 +598,36 @@ ComposerPageState.htmlToText = function(root, blacklist = []) {
 // Linkifies "plain text" link
 ComposerPageState.linkify = function(node) {
     if (node.nodeType == Node.TEXT_NODE) {
-        // Examine text node for something that looks like a URL
-        let input = node.nodeValue;
-        if (input != null) {
-            let output = input.replace(ComposerPageState.URL_REGEX, function(url) {
-                if (url.match(ComposerPageState.PROTOCOL_REGEX) != null) {
-                    url = "\x01" + url + "\x01";
+        while (node.nodeValue) {
+            // Examine text node for something that looks like a URL
+            let urlRegex = new RegExp(ComposerPageState.URL_REGEX);
+            let url;
+            do {
+                let urlMatch = urlRegex.exec(node.nodeValue);
+                if (!urlMatch) {
+                    return;
                 }
-                return url;
-            });
+                url = urlMatch[0];
+            } while (!url.match(ComposerPageState.PROTOCOL_REGEX));
 
-            if (input != output) {
-                // We got one!  Now split the text and swap in a new anchor.
-                let parent = node.parentNode;
-                let sibling = node.nextSibling;
-                for (let part of output.split("\x01")) {
-                    let newNode = null;
-                    if (part.match(ComposerPageState.URL_REGEX) != null) {
-                        newNode = document.createElement("A");
-                        newNode.href = part;
-                        newNode.innerText = part;
-                    } else {
-                        newNode = document.createTextNode(part);
-                    }
-                    parent.insertBefore(newNode, sibling);
-                }
-                parent.removeChild(node);
-            }
+            // We got one! Now split the text and swap in a new anchor.
+            let before = node.nodeValue.substring(0, urlRegex.lastIndex - url.length);
+            let after = node.nodeValue.substring(urlRegex.lastIndex);
+
+            let beforeNode = document.createTextNode(before);
+            let linkNode = document.createElement("A");
+            linkNode.href = url;
+            linkNode.textContent = url;
+            let afterNode = document.createTextNode(after);
+
+            let parentNode = node.parentNode;
+            parentNode.insertBefore(beforeNode, node);
+            parentNode.insertBefore(linkNode, node);
+            parentNode.insertBefore(afterNode, node);
+            parentNode.removeChild(node);
+
+            // Keep searching for URLs after this one
+            node = afterNode;
         }
     } else {
         // Recurse
